// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the GNU General Public License version 2 ("the License").
// You may not use this file except in compliance with the License.

// FIXME
package org.ibex.graphics;
import java.util.*;
import org.ibex.util.*;

/** an abstract path; may contain splines and arcs */
public class Path {

    public static final float PX_PER_INCH = 72;
    public static final float INCHES_PER_CM = (float)0.3937;
    public static final float INCHES_PER_MM = INCHES_PER_CM / 10;
    private static final int DEFAULT_PATHLEN = 1000;
    private static final float PI = (float)Math.PI;

    // the number of vertices on this path
    int numvertices = 0;

    // the vertices of the path
    float[] x = new float[DEFAULT_PATHLEN];
    float[] y = new float[DEFAULT_PATHLEN];

    // the type of each edge; type[i] is the type of the edge from x[i],y[i] to x[i+1],y[i+1]
    byte[] type = new byte[DEFAULT_PATHLEN];

    // bezier control points
    float[] c1x = new float[DEFAULT_PATHLEN];  // or rx (arcto)
    float[] c1y = new float[DEFAULT_PATHLEN];  // or ry (arcto)
    float[] c2x = new float[DEFAULT_PATHLEN];  // or x-axis-rotation (arcto)
    float[] c2y = new float[DEFAULT_PATHLEN];  // or large-arc << 1 | sweep (arcto)

    boolean closed = false;

    static final byte TYPE_MOVETO = 0;
    static final byte TYPE_LINETO = 1;
    static final byte TYPE_ARCTO = 2;
    static final byte TYPE_CUBIC = 3;
    static final byte TYPE_QUADRADIC = 4;

    // FIXME: hack
    private String toString;
    public String toString() { return toString; }

    public Path(String s) {
        this.toString = s;
        Tokenizer t = new Tokenizer(s);
        char last_command = 'M';
        boolean first = true;
        while(t.hasMoreTokens()) {
            char command = t.parseCommand();
            if (first && command == 'm') command = 'M';
            if (first && command != 'M') throw new RuntimeException("the first command of a path must be 'M'");
            first = false;
            boolean relative = Character.toLowerCase(command) == command;
            command = Character.toLowerCase(command);
            parseSingleCommandAndArguments(t, command, relative);
            last_command = command;
        }
    }

    public long transform(Affine a, boolean forReal) { return transform(a, forReal, true); }
    public long transform(Affine a, boolean forReal, boolean widthheight) {
        float minx = Integer.MAX_VALUE; float miny = Integer.MAX_VALUE;
        float maxx = Integer.MIN_VALUE; float maxy = Integer.MIN_VALUE;
        for(int i=0; i<numvertices; i++) {
            if (type[i] == TYPE_ARCTO) { /* FIXME!!! WRONG!!!! */ continue; }
            float x   = a.multiply_px(this.x[i],   this.y[i]);    if (x>maxx) maxx = x; if (x<minx) minx = x;
            float y   = a.multiply_py(this.x[i],   this.y[i]);    if (y>maxy) maxy = y; if (y<miny) miny = y;
            float c1x = a.multiply_px(this.c1x[i], this.c1y[i]);  if (c1x>maxx) maxx = c1x; if (c1x<minx) minx = c1x;
            float c1y = a.multiply_py(this.c1x[i], this.c1y[i]);  if (c1y>maxy) maxy = c1y; if (c1y<miny) miny = c1y;
            float c2x = a.multiply_px(this.c2x[i], this.c2y[i]);  if (c2x>maxx) maxx = c2x; if (c2x<minx) minx = c2x;
            float c2y = a.multiply_py(this.c2x[i], this.c2y[i]);  if (c2y>maxy) maxy = c2y; if (c2y<miny) miny = c2y;
            if (forReal) {
                this.x[i] = x; this.y[i] = y;
                this.c1x[i] = c1x; this.c1y[i] = c1y;
                this.c2x[i] = c2x; this.c2y[i] = c2y;
            }
        }
        if (widthheight) return ((((long)Float.floatToIntBits(maxx - minx)) << 32) | ((long)Float.floatToIntBits(maxy - miny)));
        else return ((((long)Float.floatToIntBits(minx)) << 32) | ((long)Float.floatToIntBits(miny)));
    }

    public void alignToOrigin() {
        float minx = Integer.MAX_VALUE; float miny = Integer.MAX_VALUE;
        for(int i=0; i<numvertices; i++) { if (x[i] < minx) minx = x[i]; if (y[i] < miny) miny = y[i]; }
        for(int i=0; i<numvertices; i++) {
            x[i] -= minx; y[i] -= miny;
            if (type[i] == TYPE_ARCTO) continue;
            c1x[i] -= minx; c2x[i] -= minx; c1y[i] -= miny; c2y[i] -= miny;
        }
    }


    public static class Tokenizer {
        // FIXME: check array bounds exception for improperly terminated string
        String s;
        int i = 0;
        char lastCommand = 'M';
        public Tokenizer(String s) { this.s = s; }
            
        public static Path parse(String s) {
            if (s == null) return null;
            Tokenizer t = new Tokenizer(s);
            Path ret = new Path(s);
            char last_command = 'M';
            boolean first = true;
            while(t.hasMoreTokens()) {
                char command = t.parseCommand();
                if (first && command != 'M') throw new RuntimeException("the first command of a path must be 'M'");
                first = false;
                boolean relative = Character.toLowerCase(command) == command;
                command = Character.toLowerCase(command);
                ret.parseSingleCommandAndArguments(t, command, relative);
                last_command = command;
            }
            return ret;
        }

        private void consumeWhitespace() {
            while(i < s.length() && (Character.isWhitespace(s.charAt(i)))) i++;
            if (i < s.length() && s.charAt(i) == ',') i++;
            while(i < s.length() && (Character.isWhitespace(s.charAt(i)))) i++;
        }
        public boolean hasMoreTokens() { consumeWhitespace(); return i < s.length(); }
        public char parseCommand() {
            consumeWhitespace();
            char c = s.charAt(i);
            if (!Character.isLetter(c)) return lastCommand;
            i++;
            return lastCommand = c;
        }
        public float parseFloat() {
            consumeWhitespace();
            int start = i;
            float multiplier = 1;
            for(; i < s.length(); i++) {
                char c = s.charAt(i);
                if (Character.isWhitespace(c) || c == ',' || (c == '-' && i != start)) break;
                if (!((c >= '0' && c <= '9') || c == '.' || c == 'e' || c == 'E' || c == '-')) {
                    if (c == '%') {                                // FIXME
                    } else if (s.regionMatches(i, "pt", 0, i+2)) { // FIXME
                    } else if (s.regionMatches(i, "em", 0, i+2)) { // FIXME
                    } else if (s.regionMatches(i, "pc", 0, i+2)) { // FIXME
                    } else if (s.regionMatches(i, "ex", 0, i+2)) { // FIXME
                    } else if (s.regionMatches(i, "mm", 0, i+2)) { i += 2; multiplier = INCHES_PER_MM * PX_PER_INCH; break;
                    } else if (s.regionMatches(i, "cm", 0, i+2)) { i += 2; multiplier = INCHES_PER_CM * PX_PER_INCH; break;
                    } else if (s.regionMatches(i, "in", 0, i+2)) { i += 2; multiplier = PX_PER_INCH; break;
                    } else if (s.regionMatches(i, "px", 0, i+2)) { i += 2; break;
                    } else if (Character.isLetter(c)) break;
                    throw new RuntimeException("didn't expect character \"" + c + "\" in a numeric constant");
                }
            }
            //if (start == i) throw new RuntimeException("FIXME");
            if (start == i) return (float)0.0;
            try {
                return Float.parseFloat(s.substring(start, i)) * multiplier;
            } catch (NumberFormatException nfe) {
                Log.warn(Path.class, "offending string was \"" + s.substring(start, i) + "\"");
                throw nfe;
            }
        }
    }

    /** Creates a concrete vector path transformed through the given matrix. */
    public void addTo(Polygon ret, Affine a) {
        float NUMSTEPS = 5;  // FIXME
        ret.x[0] = a.multiply_px(x[0], y[0]);
        ret.y[0] = a.multiply_py(x[0], y[0]);

        for(int i=0; i<numvertices; i++) {
            if (type[i] == TYPE_LINETO) {
                float rx = x[i+1];
                float ry = y[i+1];
                ret.add(a.multiply_px(rx, ry), a.multiply_py(rx, ry));

            } else if (type[i] == TYPE_MOVETO) {
                float rx = x[i+1];
                float ry = y[i+1];
                ret.newcontour();
                ret.add(a.multiply_px(rx, ry), a.multiply_py(rx, ry));

            } else if (type[i] == TYPE_ARCTO) {
                float rx = c1x[i];
                float ry = c1y[i];
                float phi = c2x[i];
                float fa = ((int)c2y[i]) >> 1;
                float fs = ((int)c2y[i]) & 1;
                float x1 = x[i];
                float y1 = y[i];
                float x2 = x[i+1];
                float y2 = y[i+1];

                // F.6.5: given x1,y1,x2,y2,fa,fs, compute cx,cy,theta1,dtheta
                float x1_ = (float)Math.cos(phi) * (x1 - x2) / 2 + (float)Math.sin(phi) * (y1 - y2) / 2;
                float y1_ = -1 * (float)Math.sin(phi) * (x1 - x2) / 2 + (float)Math.cos(phi) * (y1 - y2) / 2;
                float tmp = (float)Math.sqrt((rx * rx * ry * ry - rx * rx * y1_ * y1_ - ry * ry * x1_ * x1_) /
                                             (rx * rx * y1_ * y1_ + ry * ry * x1_ * x1_));
                float cx_ = (fa == fs ? -1 : 1) * tmp * (rx * y1_ / ry);
                float cy_ = (fa == fs ? -1 : 1) * -1 * tmp * (ry * x1_ / rx);
                float cx = (float)Math.cos(phi) * cx_ - (float)Math.sin(phi) * cy_ + (x1 + x2) / 2;
                float cy = (float)Math.sin(phi) * cx_ + (float)Math.cos(phi) * cy_ + (y1 + y2) / 2;

                // F.6.4 Conversion from center to endpoint parameterization
                float ux = 1, uy = 0, vx = (x1_ - cx_) / rx, vy = (y1_ - cy_) / ry;
                float det = ux * vy - uy * vx;
                float theta1 = (det < 0 ? -1 : 1) *
                    (float)Math.acos((ux * vx + uy * vy) / 
                                     ((float)Math.sqrt(ux * ux + uy * uy) * (float)Math.sqrt(vx * vx + vy * vy)));
                ux = (x1_ - cx_) / rx; uy = (y1_ - cy_) / ry;
                vx = (-1 * x1_ - cx_) / rx; vy = (-1 * y1_ - cy_) / ry;
                det = ux * vy - uy * vx;
                float dtheta = (det < 0 ? -1 : 1) *
                    (float)Math.acos((ux * vx + uy * vy) / 
                                     ((float)Math.sqrt(ux * ux + uy * uy) * (float)Math.sqrt(vx * vx + vy * vy)));
                dtheta = dtheta % (float)(2 * Math.PI);

                if (fs == 0 && dtheta > 0) theta1 -= 2 * PI; 
                if (fs == 1 && dtheta < 0) theta1 += 2 * PI;

                if (fa == 1 && dtheta < 0) dtheta = 2 * PI + dtheta;
                else if (fa == 1 && dtheta > 0) dtheta = -1 * (2 * PI - dtheta);

                // FIXME: integrate F.6.6
                // FIXME: isn't quite ending where it should...

                // F.6.3: Parameterization alternatives
                float theta = theta1;
                for(int j=0; j<NUMSTEPS; j++) {
                    float rasterx = rx * (float)Math.cos(theta) * (float)Math.cos(phi) -
                        ry * (float)Math.sin(theta) * (float)Math.sin(phi) + cx;
                    float rastery = rx * (float)Math.cos(theta) * (float)Math.sin(phi) +
                        ry * (float)Math.cos(phi) * (float)Math.sin(theta) + cy;
                    ret.add(a.multiply_px(rasterx, rastery), a.multiply_py(rasterx, rastery));
                    theta += dtheta / NUMSTEPS;
                }

            } else if (type[i] == TYPE_CUBIC) {

                float ax = x[i+1] - 3 * c2x[i] + 3 * c1x[i] - x[i];
                float bx = 3 * c2x[i] - 6 * c1x[i] + 3 * x[i];
                float cx = 3 * c1x[i] - 3 * x[i];
                float dx = x[i];
                float ay = y[i+1] - 3 * c2y[i] + 3 * c1y[i] - y[i];
                float by = 3 * c2y[i] - 6 * c1y[i] + 3 * y[i];
                float cy = 3 * c1y[i] - 3 * y[i];
                float dy = y[i];
		    
                float x0 = a.multiply_px(x[i], y[i]);
                float y0 = a.multiply_py(x[i], y[i]);
                float x1 = a.multiply_px(x[i+1], y[i+1]);
                float y1 = a.multiply_py(x[i+1], y[i+1]);
                float steps = (float)Math.sqrt( (x1-x0) * (x1-x0) + (y1-y0) * (y1-y0) );

                for(float t=0; t<1; t += 1 / (steps/20)) {
                    float rx = ax * t * t * t + bx * t * t + cx * t + dx;
                    float ry = ay * t * t * t + by * t * t + cy * t + dy;
                    ret.add(a.multiply_px(rx, ry), a.multiply_py(rx, ry));
                }


            } else if (type[i] == TYPE_QUADRADIC) {

                float bx = x[i+1] - 2 * c1x[i] + x[i];
                float cx = 2 * c1x[i] - 2 * x[i];
                float dx = x[i];
                float by = y[i+1] - 2 * c1y[i] + y[i];
                float cy = 2 * c1y[i] - 2 * y[i];
                float dy = y[i];
		    	
                float x0 = a.multiply_px(x[i], y[i]);
                float y0 = a.multiply_py(x[i], y[i]);
                float x1 = a.multiply_px(x[i+1], y[i+1]);
                float y1 = a.multiply_py(x[i+1], y[i+1]);
                float steps = (float)Math.sqrt( (x1-x0) * (x1-x0) + (y1-y0) * (y1-y0) );

                for(float t=0; t<1; t += 1 / (steps/20)) {
                    float rx = bx * t * t + cx * t + dx;
                    float ry = by * t * t + cy * t + dy;
                    ret.add(a.multiply_px(rx, ry), a.multiply_py(rx, ry));
                }

            }
        }
    }

    protected void parseSingleCommandAndArguments(Tokenizer t, char command, boolean relative) {
        if (numvertices == 0 && command != 'm')
            throw new RuntimeException("first command MUST be an 'm', not a " + command);
        if (numvertices > x.length - 2) {
            float[] new_x = new float[x.length * 2]; System.arraycopy(x, 0, new_x, 0, x.length); x = new_x;
            float[] new_y = new float[y.length * 2]; System.arraycopy(y, 0, new_y, 0, y.length); y = new_y;
        }
        switch(command) {
            case 'z': {
		int where;
                type[numvertices-1] = TYPE_LINETO;
		for(where = numvertices-2; where >= 0 && type[where] != TYPE_MOVETO; where--);
                x[numvertices] = x[where+1];
                y[numvertices] = y[where+1];
                numvertices++;
                closed = true;
                // FIXME: actually, we should search back to the last 'z' or 'm', not just 'm'
                break;
            }

            case 'm': {
                if (numvertices > 0) type[numvertices-1] = TYPE_MOVETO;
		x[numvertices] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
		y[numvertices] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
                if (numvertices > 2 && type[numvertices-2] == TYPE_MOVETO) {
                    x[numvertices-1] = x[numvertices];
                    y[numvertices-1] = y[numvertices];
                } else {
                    numvertices++;
                }
                break;
            }

            case 'l': case 'h': case 'v': {
                type[numvertices-1] = TYPE_LINETO;
                float first = t.parseFloat(), second;
                if (command == 'h') {
                    second = relative ? 0 : y[numvertices - 1];
                } else if (command == 'v') {
                    second = first; first = relative ? 0 : x[numvertices - 1];
                } else {
                    second = t.parseFloat();
                }
                x[numvertices] = first + (relative ? x[numvertices - 1] : 0);
                y[numvertices] = second + (relative ? y[numvertices - 1] : 0);
                numvertices++;
                break;
            }
            
            case 'a': {
		type[numvertices-1] = TYPE_ARCTO;
		c1x[numvertices-1] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
		c1y[numvertices-1] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
		c2x[numvertices-1] = (t.parseFloat() / 360) * 2 * PI;
		c2y[numvertices-1] = (((int)t.parseFloat()) << 1) | (int)t.parseFloat();
		x[numvertices] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
		y[numvertices] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
                numvertices++;
                break;
            }

            case 's': case 'c': {
                type[numvertices-1] = TYPE_CUBIC;
                if (command == 'c') {
                    c1x[numvertices-1] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
                    c1y[numvertices-1] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
                } else if (numvertices > 1 && type[numvertices-2] == TYPE_CUBIC) {
                    c1x[numvertices-1] = 2 * x[numvertices - 1] - c2x[numvertices-2];
                    c1y[numvertices-1] = 2 * y[numvertices - 1] - c2y[numvertices-2];
                } else {
                    c1x[numvertices-1] = x[numvertices-1];
                    c1y[numvertices-1] = y[numvertices-1];
                }
                c2x[numvertices-1] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
                c2y[numvertices-1] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
		x[numvertices] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
		y[numvertices] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
                numvertices++;
                break;
            }

            case 't': case 'q': {
                type[numvertices-1] = TYPE_QUADRADIC;
                if (command == 'q') {
                    c1x[numvertices-1] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
                    c1y[numvertices-1] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
                } else if (numvertices > 1 && type[numvertices-2] == TYPE_QUADRADIC) {
                    c1x[numvertices-1] = 2 * x[numvertices - 1] - c1x[numvertices-2];
                    c1y[numvertices-1] = 2 * y[numvertices - 1] - c1y[numvertices-2];
                } else {
                    c1x[numvertices-1] = x[numvertices-1];
                    c1y[numvertices-1] = y[numvertices-1];
                }
                x[numvertices] = t.parseFloat() + (relative ? x[numvertices - 1] : 0);
                y[numvertices] = t.parseFloat() + (relative ? y[numvertices - 1] : 0);
                numvertices++;
                break;
            }

            default:
                // FIXME
        }

        /*
        // invariant: after this loop, no two lines intersect other than at a vertex
        // FIXME: cleanup
        int index = numvertices - 2;
        for(int i=0; i<Math.min(numvertices - 3, index); i++) {
        for(int j = index; j < numvertices - 1; j++) {

        // I'm not sure how to deal with vertical lines...
        if (x[i+1] == x[i] || x[j+1] == x[j]) continue;
			
        float islope = (y[i+1] - y[i]) / (x[i+1] - x[i]);
        float jslope = (y[j+1] - y[j]) / (x[j+1] - x[j]);
        if (islope == jslope) continue;   // parallel lines can't intersect
			
        float _x = (islope * x[i] - jslope * x[j] + y[j] - y[i]) / (islope - jslope);
        float _y = islope * (_x - x[i]) + y[i];
			
        if (_x > Math.min(x[i+1], x[i]) && _x < Math.max(x[i+1], x[i]) &&
        _x > Math.min(x[j+1], x[j]) && _x < Math.max(x[j+1], x[j])) {
        // FIXME: something's not right in here.  See if we can do without fracturing line 'i'.
        for(int k = ++numvertices; k>i; k--) { x[k] = x[k - 1]; y[k] = y[k - 1]; }
        x[i+1] = _x;
        y[i+1] = _y;
        x[numvertices] = x[numvertices - 1];  x[numvertices - 1] = _x;
        y[numvertices] = y[numvertices - 1];  y[numvertices - 1] = _y;
        edges[numedges++] = numvertices - 1; numvertices++;
        index++;
        break;  // actually 'continue' the outermost loop
        }
        }
        }
        */

    }

}
